'-------------------------------------------------------------------------------
'name                     : AVRcntrlSi570
'copyright                : (c) 2008, Dipl.Ing. Dr. Christian Hirt, Austria
'purpose                  : Si570 Control with double precision
'micro                    : ATmega8515-16
'pxo                      : Si570 with Part Number 570BBC000141DG
'flash                    : 88% used
'programmer               : homebrew (STK200/STK300)
'compiler                 : MCS BASCOM-AVR 1.11.9.1
'build                    : 12062008 (12. June 2008)
'status                   : tested, but without warranty!
'hardware                 : AVR + Si570 + RotaryEncoder + LCD
'-------------------------------------------------------------------------------

$regfile = "m8515.dat"                  'ATmega8515
'$PROG &hFF , &hE1 , &hD9 , &h00        'Fuse bits. Take care !!
$crystal = 8000000                      '8 MHz
'stack and framesize not optimized!
$hwstack = 128
$framesize = 128
$swstack = 128
$lib "double.lbx"

'--- Config I2C ----------------
Config SDA = PortD.6
Config SCL = PortD.7

'--- Config LCD ----------------
Config LCD = 16 * 2
Config LCDmode = PORT
Config LCDpin = Pin , RS = PortA.0 , E = PortA.1
Config LCDpin = Pin , DB4 = PortA.2 , DB5 = PortA.3 , DB6 = PortA.4 , DB7 = PortA.5
Config LCDbus = 4

'--- Config Port and Interrupt -----
Config PIND.4 = Input
Set PIND.4
Config Int0 = Falling
On INT0 isr_RotEncoder                  'Interrupt-Routine Extern Int0

'--- Si570 data ----------------
'Const AddrSi570 = &h55                 'SI570 I2C slave base address
Const Si570WAdr = &hAA                  'I2C write mode
Const Si570RAdr = &hAB                  'I2C read  mode
Const SUF = 56321540.0                  'StartUpFrequency [Hz]
Const P2_28 = 268435456                 '2^28
Const FOSCmin = 4736328                 'FOSCmin / 1024
Const FOSCmax = 5537109                 'FOSCmax / 1024

'------------------------------------------
Declare Sub ReadStartupConfig()
Declare Sub CalculateNewConfig(ByVal Freq As Long)
Declare Sub SetNewFrequency(ByVal Freq_out As Long)
Declare Sub Write_Si570(Byval Addr As Byte , Byval Value As Byte)
Declare Function Read_Si570(Byval Addr As Byte) As Byte
Declare Sub ShowFreq(ByVal F As Long)

Dim FOUT As Long                        'frequency
Dim reg(6) As Byte                      'register map
Dim HSDIVA(6) As Byte                   'valid HSDIV value array
Dim HSDIV as Long                       'HSDIV of Si570
Dim N1 as Long                          'N1 of Si570
Dim RFREQ As Double                     'RFREQ of Si570
Dim FXTAL As Double                     'XTAL frequency of Si570
Dim RotEnc As Long                      'RotaryEncoder counter
Dim RotEncOld As Long                   'RotaryEncoder counter
Dim i As Byte
Dim a As Byte
Dim dtemp As Double
Dim L as Long

'---- Main ---------------------------------------------------------------------
Enable Interrupts                       'enable global interrupts
Enable INT0                             'switch extern interrupt0 on
Cursor off
Cls                                     'clear LCD
Lcd "Si570 uC TEST"
LowerLine
Lcd "(c) Chris Hirt"
Wait 2
Lcd Spc(16)
Locate 2 , 1
'#############################################################
FOUT = 28600000                      'new start frequency [Hz]
'#############################################################
RotEnc = FOUT
RotEncOld = RotEnc
Call ReadStartupConfig()                'read Si570 startup configuration
For i = 0 to 5
   HSDIVA(i + 1) = Lookup(i , HSDIVdata)       'get HSDIVA array{11,9,7,5,6,4}
Next i
Call SetNewFrequency(FOUT)              'calculate and write new frequency
Call ShowFreq(FOUT)                     'show frequency with LCD
Do                                      'main loop
   If RotEnc <> RotEncOld Then
      L = RotEnc - RotEncOld
      L = L * 200                       'frequency increment x 2 (100 Hz)
      FOUT = FOUT + L
      If FOUT < 10000000 Then           'Si570 frequency range MIN (10 MHz)
         FOUT = 10000000
      End If
      If FOUT > 215000000 Then          'Si570 frequency range MAX (215 MHz)
         FOUT = 215000000
      End If
      RotEnc = FOUT
      RotEncOld = RotEnc
      Call SetNewFrequency(FOUT)        'calculate and write new frequency
      Call ShowFreq(FOUT)               'show frequency with LCD
   End If
Loop

'-------------------------------------------------------------------------------

Sub ReadStartupConfig()                 'read Si570 startup configuration
   Local r1 As Byte , r2 As Byte
   Local RFREQlng As Long

   r1 = Read_Si570(135)                 'read register 135
   r1 = r1 OR &b0000_0001               'set bit D0 = 1, others any state
   Call Write_Si570(135 , r1)           'recall NVM Freq  --> D0 = 1
   r1 = 1
   While r1 <> 0                        'wait until RECALL = 0
      r1 = Read_Si570(135)
      r1 = r1 And &b0000_0001
   WEnd
   For i = 1 To 6
      a = i + 6
      reg(i) = Read_Si570(a)            'read registers 7 to 12 of Si570
   Next i
   'HS_DIV conversion
   HSDIV = reg(1) And &b1110_0000
   Shift HSDIV , Right , 5
   HSDIV = HSDIV + 4                    'HS_DIV = {4,5,6,7,9,11}
   'N1 conversion
   r1 = reg(1) And &b0001_1111
   Shift r1 , Left , 2
   r2 = reg(2) And &b1100_0000
   Shift r2 , Right , 6
   N1 = r1 + r2                         'N1 = {1,2,4,6...128}
   If N1 = 0 Then
      N1 = 1
   Else
      r1 = N1 And &b0000_0001
      If r1 <> 0 Then
         N1 = N1 + 1
      End If
   End If
   'RFREQ conversion
   RFREQlng = reg(2) And &b0011_1111
   shift RFREQlng , left , 4
   r1 = reg(3) And &b1111_0000
   shift r1 , right , 4
   RFREQlng = RFREQlng + r1             'integer portion of RFREQ (10 MSB = INT)
   dtemp = RFREQlng
   RFREQlng = reg(3) And &b0000_1111
   shift RFREQlng , left , 8
   RFREQlng = RFREQlng + reg(4)
   shift RFREQlng , left , 8
   RFREQlng = RFREQlng + reg(5)
   shift RFREQlng , left , 8
   RFREQlng = RFREQlng + reg(6)         'frac portion of RFREQ (28 LSB = FRAC)
   RFREQ = RFREQlng
   RFREQ = RFREQ / P2_28
   RFREQ = dtemp + RFREQ
   'Actual nominal crystal frequency
   dtemp = HSDIV
   FXTAL = SUF * dtemp
   dtemp = N1
   FXTAL = FXTAL * dtemp
   FXTAL = FXTAL / RFREQ
End Sub

'-------------------------------------------------------------------------------

Sub CalculateNewConfig(ByVal Freq As Long)  ' calculate new frequency config
   Local N1min As Long
   Local N1max As Long
   Local FOSC As Long
   Local FOSCtmp As Long

   shift Freq , right , 10
   N1max = FOSCmax / Freq
   N1min = FOSCmin / Freq
   shift N1max , right , 2
   a = N1max mod 2                      'if a = 1 then N1max = odd
   If a = 1 then
     N1max = N1max + 1
   EndIf
   N1min = N1min / 11
   a = N1min mod 2                      'if a = 1 then N1max = odd
   If a = 1 then
      N1min = N1min - 1
   EndIf
   N1min = N1min + 1
   If N1max > 128 then N1max = 128      'N1 range {1...128}
   If N1min <= 0 then N1min = 1
   FOSCtmp = 0
   Do                                   'test all valide N1 values
      For i = 1 To 6                    'with all valide HSDIV values
         FOSC = Freq * HSDIVA(i)        'FOSC = Fnew x HS_DIV x N1
         FOSC = FOSC * N1max
         If FOSC >= FOSCmin then
            If FOSC <= FOSCmax then
               If FOSCtmp = 0 then
                  FOSCtmp = FOSC
                  HSDIV = HSDIVA(i)
                  N1 = N1max
               EndIf
               If FOSC <= FOSCtmp then
                  FOSCtmp = FOSC
                  HSDIV = HSDIVA(i)
                  N1 = N1max
               EndIf
            EndIf
         EndIf
      Next i
      N1max = N1max - 2
   Loop Until N1max <= N1min
End Sub

'-------------------------------------------------------------------------------

Sub SetNewFrequency(ByVal Freq_Out As Long)     'pack and write new dividers
   Local r3 As Long , r4 As Long , r5 As Long
   Local rb As Byte
   Local N11 As Byte , r1 As Byte , N1a As Byte , N1b As Byte
   Local dtemp2 As Double
   Local k As Long
   Local temp As Long

   Call CalculateNewConfig(Freq_Out)    '--> New HSDIV , N1
   'RFREQ = F1 x HS_DIV x N1 / FXTAL
   dtemp = HSDIV
   dtemp2 = Freq_Out
   RFREQ = dtemp * dtemp2
   dtemp = N1
   RFREQ = RFREQ * dtemp
   RFREQ = RFREQ / FXTAL                ' --> New RFREQ
   '-------------- Pack Si570 combo -------------
   For i = 1 To 6
      reg(i) = 0                        'clear register map
   Next i
   'new HS_DIV conversion
   HSDIV = HSDIV - 4
   Shift HSDIV , Left , 5
   reg(1) = HSDIV                       'set register 7 bits 5, 6, 7
   'new N1 conversion
   If N1 = 1 Then
      N1 = 0
   Else
      N11 = N1 And 1
      If N11 = 0 Then
         N1 = N1 - 1                    'if N1 is even, subtract one
      End If
   End If
   r1 = reg(1) And &b1110_0000
   N1a = N1
   Shift N1a , Right , 2
   reg(1) = r1 Or N1a                   'set register 7 bits 0 to 4
   N1b = N1 And &b0000_0011
   Shift N1b , Left , 6
   reg(2) = N1b                         'set register 8 bits 6 and 7
   'new RFREQ conversion
   dtemp = Frac(RFREQ)                  'FRAC portion
   dtemp2 = RFREQ - dtemp               'INT Portion
   k = dtemp2
   shift k , right , 4
   rb = k AND &b0011_1111
   reg(2) = reg(2) OR rb                'set register 8 bits 0 to 5
   dtemp = dtemp * P2_28                ' x 2^28
   temp = dtemp
   r3 = temp
   shift r3 , right , 24
   reg(3) = r3 And &b1111_1111
   k = dtemp2
   shift k , left , 4
   reg(3) = reg(3) OR k
   r4 = temp
   shift r4 , right , 16
   reg(4) = r4 And &b1111_1111
   r5 = temp
   shift r5 , right , 8
   reg(5) = r5 And &b1111_1111
   reg(6) = temp And &b1111_1111
   '-------------- Write Si570 registers ------------
   rb = Read_Si570(137)
   rb = rb OR &b0001_0000               'set bit 4
   Call Write_Si570(137 , rb)           'freeze DCO !
   For i = 1 To 6
      a = i + 6                         'modify the frequency configuration
      Call Write_Si570(a , reg(i))
   Next a
   rb = Read_Si570(137)
   rb = rb And &b1110_1111              'reset bit 4
   Call Write_Si570(137 , rb)           'unfreeze DCO !
   rb = Read_Si570(135)
   rb = rb OR &b0100_0000               'set bit 6 (NewFreq)
   Call Write_Si570(135 , rb)           'new freq applied
   rb = 1
   While rb <> 0                        'wait until NewFreq = 0
      rb = Read_Si570(135)
      rb = rb And &b0100_0000
   WEnd
End Sub

'-------------------------------------------------------------------------------

Sub ShowFreq(ByVal F As Long)           'show frequency with LCD
  Local s as String * 9
  Local s1 as String * 11
  F = F / 2                             'QLO divider  (2 , 4 , ..)
  s = Str(F)
  s = Format(s , "000000000")
  s1 = Left(s , 3) + "." + Mid(s , 4 , 3) + "." + Right(s , 3)
  LowerLine
  Lcd Spc(16)
  Locate 2 , 1
  Lcd "F " ; s1 ; " Hz"
End Sub

'--- I2C functions -------------------------------------------------------------

Sub Write_Si570(byval Addr As Byte , Byval Value As Byte)       'write a byte
   I2cstart                             'start condition
   I2cwbyte Si570WAdr                   'write mode address from Si570
   I2cwbyte Addr                        'adress of Si570 register
   I2cwbyte Value                       'byte value to write
   I2cstop                              'stop condition
End Sub

Function Read_Si570(byval Addr As Byte) As Byte       'read a byte from Si570
  Dim Rdval As Byte
  Rdval = 0
  I2cstart                              'start condition
  I2cwbyte Si570WAdr                    'write mode adress from Si570
  I2cwbyte Addr                         'address of Si570 register
  I2cstart                              'repeated start
  I2cwbyte Si570RAdr                    'read mode address from Si570
  I2crbyte Rdval , Nack                 'read byte
  I2cstop                               'generate stop
  Read_Si570 = Rdval
End Function

'----- ISR Read Encoder --------------------------------------------------------

isr_RotEncoder:                         'extern interrupt0 routine
   If PIND.4 = 0 Then
      Incr RotEnc
   Else
      Decr RotEnc
   End If
   Gifr = 64
Return

'--- Data table ----------------------------------------------------------------

HSDIVdata:
   Data 4 , 5 , 6 , 7 , 9 , 11

'------------------------- End of AVRcntrlSi570 --------------------------------